// (c) Meta Platforms, Inc. and affiliates. Confidential and proprietary.

#pragma once

#include <string>

namespace KINETO_NAMESPACE {

std::string demangle(const char* name);
std::string demangle(const std::string& name);

} // namespace KINETO_NAMESPACE
